/**
 * Copyright <2022> <Innovative Sensor Technology IST AG>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * 4. This license is granted under the laws of Switzerland, and the Courts of
 * Basle City shall have exclusive jurisdiction for any dispute arising in the
 * context of this license.  
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (EXCLUDING GROSS NEGLIGENCE AND INTENT
 * AS WELL AS OTHER MANDATORY LEGAL PROVISONS SUCH AS PRODUCT LIABILITY LAWS)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE.
 */

//
//Projekt:		Software I2C-Master
//Datei:		main.c
//Datum:		2011-11-09
//Autor:		Mike Steitz
//
// Software I2C-Master zur Abarbeitung eines HumiPicco 27 der IST AG
// Programm lauffaehig auf einem ATMEL AtTiny 84 unter AVR-Studio Version 4.18 Build 716
// der Anschluss des Sensors erfolgt ueber die Ports PA2 (SCL) und PA3 (SDA)
//


#include <avr/io.h>
#include <avr/interrupt.h>


#include "timer.h"
#include "sw_master.h"

#define SEND_             0x01
#define REQUEST_          0x02
#define I2C_BUSY_		  0x04

static unsigned char I2C_Act_On_Failure_In_Last_Transmission ( unsigned char errorcode )
{
  unsigned char rval;
  unsigned char next_state;
  // Fehlerbehandlung fr nicht erfolgreichen Treiberaufruf

  if (errorcode & I2C_SCL_STAY_LOW || errorcode & I2C_SDA_STAY_LOW ) {
    rval = init_i2c_master();
	if (rval == I2C_OK) {
	  next_state = SEND_;
	} else {
	  next_state = I2C_BUSY_;
	}
  } else {}


  return next_state;
}


float		temperature;
float		humidity;
uint8_t   	stale_bit;

int main(void) {
  uint8_t   	msg[4];
  uint8_t   	rval;
  uint8_t   	slaveAddress; 
  unsigned int 	raw_value_temp;
  unsigned int	raw_value_humid;
  uint8_t   	i2c_status;
  uint8_t   	df_not_ready = 0;



  cli(); // disable interrupt

  timer1_init();
  rval = init_i2c_master();
  if (rval == I2C_OK) {
    i2c_status = SEND_;
  } else {
    i2c_status = I2C_BUSY_;
  }

  // I2C-Adresse des Slaves
  slaveAddress = 0x28;

  sei(); // enable_interrupt


  for(;;) { 

    // Zustand wechselt stndig zwischen senden und empfangen
	switch (i2c_status ) {
	  case SEND_:
	    // im Zustand senden wird das MR-Command an den HumiPicco gesendet
 	    // das MR-Command besteht aus der Uebertragung der Adresse ohne Daten
	    rval = i2c_master_write(slaveAddress, &msg[0], 0); 
	    if (rval == I2C_OK) {
          
          i2c_status = REQUEST_; // auf Empfang umschalten
		  df_not_ready = 0;		   // DF-Command not ready counter
		
		  // Verzgerung zwischen dem MR- und DF- Command, um dem Sensor Zeit fuer
		  // die Datenermittlung zu geben, hier wird 100ms gewartet
		  i2c_time_out_timer = 0;
		  while (i2c_time_out_timer < 10) {
		  }

        } else {
          // Errorhandling, falls die Uebertragung des MR-Comands fehlschlug
		  i2c_status = I2C_BUSY_;
	    }
		break;
	  case REQUEST_:
	    // im Zustand empfangen werden das DF-Command ausgeloest
	    // das DF-Command besteht aus dem auslesen von 4 Datenbytes
        rval = i2c_master_read(slaveAddress, &msg[0], 4);
      
	    if (rval == I2C_OK) {
	      // die beiden hoechstwertigsten Bits des ersten Datenbytes enthalten Statusinformationen
		  // das Bit 0x40 ist das Stale-Bit, ist es 1 sind noch keine neuen Daten vorhanden
  	      stale_bit = (msg[0] & 0x40) >> 6;
	      if (stale_bit == 0) {
            // Rohwert der Temperatur als 14bit-Wert im 3. und 4. Datenbyte, der Wert ist linksbuendig 
		    // ausgerichtet, deshalb >> 2
 	        raw_value_temp = ((msg[2] << 8) | msg[3]) >> 2;
		    // Rohwert der Feuchte als 14bit -Wert im 1. und 2. Datenbyte, der Wert ist rechtsbuendig
		    // ausgerichtet, im 1. Datenbyte mussen die beiden Statusbits geloescht werden
  		    raw_value_humid = ((msg[0] & 0x3F) << 8) | msg[1];
  
 		    if (raw_value_temp < 0x3FFF && raw_value_humid < 0x3FFF) {
              //Umrechnung der Rohwerte in die Nutzdaten
		      // Temperatur als 14bit-Wert im Bereich von -40C - +125C
		      temperature = ((float)(raw_value_temp) * 165.0F / 16383.0F) - 40.0F;
		      // Feuchte als 14bit-Wert im Bereich von 0% - 100%
              humidity = (float)raw_value_humid * 100.0F / 16383.0F;
  		    } else {
			  // Fehlerbehandlung, der Sensor signalisiert durch Endanschlag der beiden
			  // Messwerte einen internen Fehler
		    }
            i2c_status = SEND_;    // Set next operation        
	      } else {
	        // Daten noch nicht fertig, hier wird noch einmal gewartet und erneut gelesen
		    // ist danach noch keine Antwort da, dann wir MR erneut gesendet
	      
		    i2c_time_out_timer = 0;
		    while (i2c_time_out_timer < 10) {
		    }
		    df_not_ready++;
		    if (df_not_ready > 1) {
              i2c_status = SEND_;
		    } else {
    	      i2c_status = REQUEST_;
            }
	      }
        } else {
          // Errorhandling, falls die Uebertragung des DF-Comands fehlschlug
          i2c_status = I2C_BUSY_;    // Set next operation        
      	}
		break;
	  case I2C_BUSY_:
		i2c_status = I2C_Act_On_Failure_In_Last_Transmission (i2c_errorstate);
        break;
    }
   


  }
}



