/**
 * Copyright <2022> <Innovative Sensor Technology IST AG>
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from this
 * software without specific prior written permission.
 * 
 * 4. This license is granted under the laws of Switzerland, and the Courts of
 * Basle City shall have exclusive jurisdiction for any dispute arising in the
 * context of this license.  
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (EXCLUDING GROSS NEGLIGENCE AND INTENT
 * AS WELL AS OTHER MANDATORY LEGAL PROVISONS SUCH AS PRODUCT LIABILITY LAWS)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE.
 */

//
//Projekt:		Software I2C-Master
//Datei:		SW_master.c
//Datum:		2011-11-09
//Autor:		Mike Steitz
//

#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>


#include "timer.h"

#define extern
#include "sw_master.h"
#undef extern


//****************************************************************************
//
// Prototypen lokaler Funktionen
//
//****************************************************************************
static void sda_low(void);
static void sda_high(void);
static void scl_low(void);
static void scl_high(void);
static unsigned char i2c_try_scl(void);
static unsigned char i2c_m_clk_toggle(void);
static unsigned char i2c_m_start(void);
static unsigned char i2c_m_stop(void);
static unsigned char i2c_m_write(unsigned char b);
static unsigned char i2c_m_read(unsigned char *read_byte, unsigned char a);



#define I2C_SCL_LOW		scl_low()
#define I2C_SCL_HIGH	scl_high()

#define I2C_SDA_LOW		sda_low()
#define I2C_SDA_HIGH	sda_high()

#define	ACK				0							// Ack empfangen
#define	NACK			1							// Ack nicht empfangen

#define S_DELAY			_delay_loop_1(1);		// Delayfunktionen aus delay_basic.h
#define M_DELAY			_delay_loop_1(3);


// Initialisierung der Schnittstelle
unsigned char init_i2c_master(void) {
  unsigned char rval;

  i2c_errorstate = 0;
  I2C_SCL_HIGH;								// Cockleitung auf H setzen
  I2C_SDA_HIGH;								// Datenleitung auf H setzen
  if(!(SDA_PIN_ & (SDA_MASK_))) {
    i2c_errorstate |= I2C_SDA_STAY_LOW;
  } else {}
  if(!(SCL_PIN_ & (SCL_MASK_))) {
    i2c_errorstate |= I2C_SCL_STAY_LOW;
  } else {}

  rval = i2c_m_stop();
  if (rval != I2C_OK || i2c_errorstate > 0) {
    return I2C_ERROR;
  } else {
    return I2C_OK;
  }
}

// Nachricht auf dem Bus ausgeben
//      adr: 		Busadresse
//	    msg: 		zu uebertragende Daten
//      anz_byte: 	Anzahl der Datenbytes
//
// return: Fehlercode I2C_ERROR im Fehlerfall, sonst I2C_OK
unsigned char i2c_master_write(unsigned char adr, unsigned char *msg, unsigned char anz_byte) {
  unsigned char i;

  i2c_errorstate = 0;
  if (i2c_m_start()) {
    if (i2c_m_write(adr << 1) == I2C_OK_ACK) {
      for (i=0; i<anz_byte; i++) {
        if (i2c_m_write(*msg) == I2C_OK_ACK) {
	        msg++;
        } else {
          i2c_m_stop();
          return I2C_ERROR;
        }
      }
      if (i2c_m_stop() == I2C_OK) {
	    if (i2c_errorstate > 0) {
		  return I2C_ERROR;
		} else {
          return I2C_OK;
		}
	  } else {
        return I2C_ERROR;
	  }
    } else {
      i2c_m_stop();
      return I2C_ERROR;
    }
  } else {
    return I2C_ERROR;
  }

}

// Nachricht auf dem Bus empfangen
// 	  	adr: 			Busadresse
//	    recv: 			zu uebertragende Daten
//      anz_byte_read: 	Anzahl der Datenbytes
//
// return: Fehlercode I2C_ERROR im Fehlerfall, sonst I2C_OK
unsigned char i2c_master_read(unsigned char adr, unsigned char *recv, unsigned char anz_byte_read) {
  unsigned char i;
  unsigned char error = 0;
  unsigned char read_byte;
  unsigned char rval;

  i2c_errorstate = 0;
  if (i2c_m_start()) {
    if (i2c_m_write((adr << 1) | 0x01) == I2C_OK_ACK) {
      for (i=0; i<anz_byte_read; i++) {
	    if (i==(anz_byte_read-1)) {
          rval = i2c_m_read(&read_byte, NACK);
	    } else {
          rval = i2c_m_read(&read_byte, ACK);
	    }
		*recv = read_byte;
	    recv++;
		if (rval != I2C_OK) {
		  error++;
		} else {}
      }
      if (i2c_m_stop() != I2C_OK) {
        error++;
      } else {}
    } else {
      i2c_m_stop();
      error++;
    }

  	if (error > 0 || i2c_errorstate > 0) {
	  return I2C_ERROR;
    } else {
      return I2C_OK;
	}
  } else {
    return I2C_ERROR;
  }

}


// Startbedingung fuer den I2C erzeugen
static unsigned char i2c_m_start(void) {
  I2C_SDA_HIGH;			// SDA vorsichtshalber auf HIGH ziehen
  I2C_SCL_HIGH;			// SCL auf High	
  // bergang von H->L auf SDA, wobei SCL H ist.
  if (i2c_try_scl()) {
    I2C_SDA_LOW;			// SDA auf Low ziehen
    I2C_SCL_LOW;			// SCL auf low
	return I2C_OK; 
  } else {
    i2c_errorstate = I2C_SCL_STAY_LOW;
    return I2C_ERROR;
  }
}

// Stopbedingung fuer den I2C erzeugen
static unsigned char i2c_m_stop(void) {
  I2C_SDA_LOW;			// SDA vorsichtshalber auf LOW ziehen
  I2C_SCL_HIGH;			// SCL auf High setzen
  // bergang von L->H auf SDA, wobei SCL H ist.
  if (i2c_try_scl()) {
    I2C_SDA_HIGH;			// SDA von L auf H ziehen
	return I2C_OK;
  } else {
    i2c_errorstate = I2C_SCL_STAY_LOW;
    return I2C_ERROR;
  }
}

// Senden eines Bytes
static unsigned char i2c_m_write(unsigned char b) {	
  unsigned char rval = I2C_OK;
  char i=0;
  
  for(i=0;i<8;i++) {
    if (b & 0x80) {
      I2C_SDA_HIGH;
    } else {
      I2C_SDA_LOW;
    }
    b<<=1;
    i2c_m_clk_toggle();
  }

  // Auf ACK vom Slave Prfen   
  I2C_SDA_HIGH;
  I2C_SCL_HIGH;
  if (i2c_try_scl()) {
    if (!(SDA_PIN_ & SDA_MASK_)) {
      rval = I2C_OK_ACK;
    } else {}
  } else {
    i2c_errorstate = I2C_SCL_STAY_LOW;
    rval = I2C_ERROR;
  }
  I2C_SCL_LOW;

  return rval;
}

// Empfangen eines Bytes
static unsigned char i2c_m_read(unsigned char *read_byte, unsigned char a) {	
  unsigned char i,c=0b00000000;
  unsigned char error = 0;
  unsigned char rval;

  I2C_SDA_HIGH;      

  for(i=0;i<8;i++) {
    c<<=1;				

    // Bitweise OR verknpfen mit dem Eingang, MSB zuerst
	if(SDA_PIN_ & SDA_MASK_) c|=1;			

    rval = i2c_m_clk_toggle();
	if (rval != I2C_OK) {
	  error++;
	} else {}
  }
  *read_byte = c;
	
  // ACK oder NACK | ACK=0; NACK=1
  if(a==ACK) {
    I2C_SDA_LOW;
  } else {}

  rval = i2c_m_clk_toggle();
  if (rval != I2C_OK) {
	error++;
  } else {}

  if (error > 0) {
  	return I2C_ERROR;
  } else {
  	return I2C_OK;
  }
}

//Funktionen zur Portsteuerung fuer SDA und SCL Leitung
static void sda_low(void) {
	SDA_PORT_ 		&=  ~(SDA_MASK_);    	// internen Pull-Up aus
	SDA_DDR_ 		|=   SDA_MASK_;         // Pin von SDA als Ausgang

}
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
static void sda_high(void) {
	SDA_DDR_ 		&= ~(SDA_MASK_);  	    // Pin von SDA als Eingang 
	SDA_PORT_ 		|=  SDA_MASK_;  	    // internen Pull-Up an 
} 

static void scl_low(void) {
	SCL_PORT_ 		&=  ~(SCL_MASK_);    	// internen Pull-Up aus
	SCL_DDR_ 		|=   SCL_MASK_;         // Pin von SDA als Ausgang

}
/*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~*/
static void scl_high(void) {
	SCL_DDR_ 		&= ~(SCL_MASK_);  	    // Pin von SDA als Eingang 
	SCL_PORT_ 		|=  SCL_MASK_;  	    // internen Pull-Up an 
} 

// pruefen ob die SCL-Leitung vom Slave freigegeben ist
// per Clockstretching kann der Slave die Leitung auf LOW halten
// in diesem Fall wird nach 200ms auf Fehler erkannt
static unsigned char i2c_try_scl(void) {
  i2c_time_out_timer = 0;
  // 200 ms warten, wenn SCL auf LOW gehalten wird
  while((!(SCL_PIN_ & SCL_MASK_)) && i2c_time_out_timer < 20) {
    S_DELAY;
  }
  if(!(SCL_PIN_ & (SCL_MASK_))) {
    return I2C_ERROR;
  } else {
    return I2C_OK;
  }
}

// Funktion zum Toggeln der SCL-Leitung
// ueber den Funktionsaufruf S_DELAY kann die Geschwindigkeit
// angepasst werden
// der AtTiny84 mit 8MHz kommt ohne delay auf maximal ca. 150kHz
static unsigned char i2c_m_clk_toggle(void) {

  I2C_SCL_HIGH;
  if (i2c_try_scl()) {
    
	S_DELAY;
    
	I2C_SCL_LOW;
  	return I2C_OK;
  } else {
    i2c_errorstate |= I2C_SCL_STAY_LOW;
    return I2C_ERROR;
  }
}


